module R = Rewriting;;
module U = Util;;

module SetS = Set.Make(U.Int);;
module MapF = Map.Make(R.Function);;
module MapV = Map.Make(R.Variable);;
module MapS = Map.Make(U.Int);;

type sym    = R.Function.t;;
type var    = R.Variable.t;;
type state  = U.Int.t;;
type states = SetS.t;;
type lhs    = sym * state list;;
type rhs    = states;;
type subst  = state MapV.t;;

type aut = {
  states : states;
  finals : states;
  trans  : (state list * rhs) list MapF.t;
  eps    : states MapS.t;
};;

let empty : aut = {
  states = SetS.empty;
  finals = SetS.empty;
  trans  = MapF.empty;
  eps    = MapS.empty;
};;

let fold_trans : (sym -> state list -> states -> 'a -> 'a) -> aut -> 'a -> 'a
  = fun g a -> MapF.fold (fun f v acc ->
    List.fold_left (fun acc (ls,rs) -> g f ls rs acc) acc v) a.trans
;;

let fold_eps : (state -> states -> 'a -> 'a) -> aut -> 'a -> 'a
  = fun g a -> MapS.fold g a.eps
;;

let intersect_states : aut -> states -> aut = fun a s' -> {
  states = SetS.inter a.states s';
  finals = SetS.inter a.finals s';
  trans = MapF.filter (fun _ v -> v != []) (
    MapF.map (U.List.flat_map (fun (ls,rs) ->
      if List.for_all (fun s -> SetS.mem s s') ls then
        let rs' = SetS.inter rs s' in
        if SetS.is_empty rs' then [] else [(ls,SetS.inter rs s')]
      else
        []
    )) a.trans);
  eps = MapS.filter (fun _ v -> not (SetS.is_empty v)) (
    MapS.map (SetS.inter s') a.eps);
};;

let reachable_states : aut -> states = fun a ->
  let rec go s =
    let s' = fold_trans (fun _ ls rs s' ->
      if List.for_all (fun s -> SetS.mem s s') ls
      then SetS.union s' rs else s') a SetS.empty in
    let s'' = fold_eps (fun l rs s'' ->
      if SetS.mem l s'' then SetS.union s'' rs else s'') a s' in
    if s = s'' then s else go s'' in
  go SetS.empty
;;

let productive_states : aut -> states = fun a ->
  let rec go s =
    let s' = fold_trans (fun _ ls rs s' ->
      if not (SetS.is_empty (SetS.inter rs s')) then
        List.fold_left (U.flip SetS.add) s' ls else s') a SetS.empty in
    let s'' = fold_eps (fun l rs s'' ->
      if not (SetS.is_empty (SetS.inter rs s'')) then
        SetS.add l s'' else s'') a s' in
    if s = s'' then s else go s'' in
  go SetS.empty
;;

let reduce : aut -> aut = fun a ->
  let a' = intersect_states a (reachable_states a) in
  intersect_states a' (productive_states a')
;;

let is_empty : aut -> bool = fun a ->
  SetS.is_empty (reduce a).finals
;;

