(* Copyright 2018 Jonas Schöpf
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 *
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** MODULES ******************************************************************)
module Ass = Assignment;;
open Monad;;
open Formula;;
open Util;;

(*** FUNCTIONS ****************************************************************)
(* replace by new process module begin *)
exception Stop;;

module Parse = Parsec.MakeCharParser(Parsec.LazyCharListInput);;

(* monadic if then else *)
let mite mx ma mb = mx >>= fun x -> if x then ma else mb;;

let rat_of_string c = match String.split ~d:"/" c with
 | [n] -> (Integer.of_string n,Integer.one)
 | [n;d] -> (Integer.of_string n,Integer.of_string d)
;;

let rat_of_real_string c = match String.split ~d:"sqrt(2)" c with
 | [] -> Rational.one
 | [c] -> rat_of_string c
;;

let number_of_string v =
 (* Format.eprintf "v = %s@\n%!" v; *)
 match String.split ~d:"+" v with
  | [c] ->
   (try (uncurry Number.of_64_rat (rat_of_string c))
   with _ -> Number.of_64_real Rational.zero (rat_of_real_string c))
  | [c;d] -> Number.of_64_real (rat_of_string c) (rat_of_real_string d)
;;

let (>>) = Parse.(>>);;
let (>>=) = Parse.(>>=);;
let (<?>) = Parse.(<?>);;
let (<|>) = Parse.(<|>);;

type assign_var = Boo of bool | Num of Number.t;;

let parse_smt2_model ch : (int * assign_var) list option =
  let input = Parsec.LazyCharListInput.of_channel ch in
  let delims : char list Parse.t = Parse.many @@ Parse.oneof " \n\t\r" in
  let word : string Parse.t = (
    Parse.lex @@ Parse.many1 @@ (Parse.char '-' <|> Parse.letter) >>=
    Parse.return @.@ String.of_char_list
  ) <?> "sequence of letters"
  in

  let parse_dig : string Parse.t = (
    Parse.lex @@ Parse.many1 Parse.digit >>=
    Parse.return @.@  String.of_char_list
  ) <?> "digits"
  in

  let parse_var : string Parse.t = (
    Parse.many1 @@ Parse.letter >>
    parse_dig >>=
    Parse.return
  ) <?> "arithemtic variable"
  in

  let inner_brackets : string Parse.t = (
    Parse.char '(' >>
    (* Parse.many @@ Parse.any >> *)
    Parse.char ')' >>
    Parse.return ""
  ) <?> "inner brackets"
  in

  let parse_ass : (int * assign_var) Parse.t = (
    Parse.lex @@ Parse.char '(' >>
    Parse.sep_by1 delims (Parse.tempt parse_var <|> inner_brackets <|> parse_dig <|> Parse.tempt word) >>= fun words ->
    Parse.lex @@ Parse.char ')' >>
    let var =
      List.nth words 1
      |> int_of_string
    in
    let plen = pred @@ List.length words in
    let v = List.nth words plen in
    List.nth words (pred plen)
    |> function
      | "Int" -> Parse.return (var, Num (number_of_string v))
      | "Bool" -> Parse.return (var, Boo (bool_of_string v))
  ) <?> "assignment"
  in

  let parse_outer_brackets : (int * assign_var) list Parse.t = (
    Parse.lex @@ Parse.char '(' >>
    word >>
    delims >>
    Parse.sep_by1 delims parse_ass >>= fun ass ->
    delims >>
    Parse.lex @@ Parse.char ')' >>
    delims >>
    Parse.return ass
  ) <?> "Model for Assignment"
  in

  let parse_result : (int * assign_var) list option Parse.t =
    Parse.many_till Parse.any (Parse.tempt @@ Parse.string "sat\n" <|> Parse.string "unsat\n") >>= fun prefix ->
    let _ = Format.printf "%s" @@ String.of_char_list prefix in
    delims >>
    word >>= fun result ->
    delims >>
    match String.equal result "sat" with
      | true ->
          parse_outer_brackets >>= fun rs ->
          Parse.many @@ Parse.any >>= fun add_out ->
          let _ = Format.printf "%s" @@ String.of_char_list add_out in
          Parse.return @@ Util.Option.some rs
      | false ->
          Parse.many @@ Parse.any >>
          Parse.return None
  in

  match Parse.parse (Parse.between Parse.spaces Parse.eoi parse_result) input with
    | Left e -> close_in ch; failwith (Parsec.Error.to_string e)
    | Right x -> close_in ch; x
;;

let external_cmd tool flags f =
  let (smtfile, in_hdl) = Filename.open_temp_file "internal." ".smt2" in
  let flags = List.fold_left (fun x y -> x ^ " " ^ y) " " flags in
  let fmt = Format.formatter_of_out_channel in_hdl in
  Formula.fprintf_smt2 fmt f;
  close_out in_hdl;
  let cmd = tool ^ flags ^ " " ^ smtfile in
  Unix.open_process_in cmd
;;

let rec create_ass = function
  | [] -> Ass.empty
  | (var, Num num) :: ass -> Ass.add_a (arith var) num (create_ass ass)
  | (var, Boo b) :: ass -> Ass.add_p (P var) b (create_ass ass)
;;

let solve tool flags f =
  let (smtfile, in_hdl) = Filename.open_temp_file "internal." ".smt2" in
  let fmt = Format.formatter_of_out_channel in_hdl in
  Formula.fprintf_smt2 fmt f;
  close_out in_hdl;
  Process.cmd_output tool (flags @ [smtfile])
    (Option.map create_ass @.@ parse_smt2_model)
  |> return
;;

let rec ea a ass = match a with
 | A l         -> return (try Ass.find_a (A l) ass with | Not_found -> Number.zero)
 | C r         -> return r
 | Fresh a     -> eval_a a ass
 | Add (a,b)   -> liftM2 Number.add (eval_a a ass) (eval_a b ass)
 | Sub (a,b)   -> liftM2 Number.sub (eval_a a ass) (eval_a b ass)
 | Mul (a,b)   -> liftM2 Number.mul (eval_a a ass) (eval_a b ass)
 | Ite (x,a,b) -> mite (eval_p x ass) (eval_a a ass) (eval_a b ass)
 | Min (a,b)   -> liftM2 Number.min (eval_a a ass) (eval_a b ass)
 | Max (a,b)   -> liftM2 Number.max (eval_a a ass) (eval_a b ass)
and ep f ass = match f with
 | Top       -> return true
 | Bot       -> return false
 | P _       -> return (try Assignment.val_of f ass with | Not_found -> false)
 | Not x     -> liftM not (eval_p x ass)
 | And (x,y) -> liftM2 (&&) (eval_p x ass) (eval_p y ass)
 | Or (x,y)  -> liftM2 (||) (eval_p x ass) (eval_p y ass)
 | Iff (x,y) -> liftM2 (=) (eval_p x ass) (eval_p y ass)
 | Eq (a,b)  -> liftM2 Number.eq (eval_a a ass) (eval_a b ass)
 | Gt (a,b)  -> liftM2 Number.gt (eval_a a ass) (eval_a b ass)
 | Ge (a,b)  -> liftM2 Number.ge (eval_a a ass) (eval_a b ass)
(*
and eval_a a ass = get >>= fun s -> cache s.State.ea_tbl (flip ea ass) a
and eval_p p ass = get >>= fun s -> cache s.State.ep_tbl (flip ep ass) p
*)
and eval_a a ass = ea a ass
and eval_p p ass = ep p ass
;;

