(* Copyright 2011-2018 Bertram Felgenhauer
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;

(*** MODULES ******************************************************************)
module F = Format;;
module M = Monad;;
module P = Problem;;
module XO = XmlOutput;;

module H = Hashtbl;;

type eqn = (int * int * int) list;;
type rel = bool array array;;

(*** TYPES ********************************************************************)
type flags = {
  help: bool ref;
  nfp: bool ref;
  unc: bool ref;
  uncsl: bool ref;
  unr: bool ref;
  vrb: bool ref;
};;

type t = {
  ip : P.t;
  pnfp : bool;
  punc : bool;
  puncsl : bool;
  punr : bool;
  result : Result.t;
};;

(*** DESTRUCTORS **************************************************************)
let get_ip t = t.ip;;
let get_result t = t.result;;

(*** GLOBALS ******************************************************************)
let code = "groundcr";;
let name = "Ground Confluence Processor";;
let keywords = ["confluence";"nonconfluence"];;

let flags = {
  help = ref false;
  nfp = ref false;
  unc = ref false;
  uncsl = ref false;
  unr = ref false;
  vrb = ref false;
};;

(*** CONSTRUCTORS *************************************************************)
let make p r = {
  ip = p;
  pnfp = !(flags.nfp);
  punc = !(flags.unc);
  puncsl = !(flags.uncsl);
  punr = !(flags.unr);
  result = r;
};;

(*** GLOBALS ******************************************************************)
let init _ =
 flags.help := false;
 flags.nfp := false;
 flags.unc := false;
 flags.uncsl := false;
 flags.unr := false;
 flags.vrb := false;
;;

let comment =
 "Decides whether the given applicative, ground TRS is confluent."
;;

let spec =
 let spec = [
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-nfp",Arg.Set flags.nfp,"Decide NFP property instead of confluence.");
  ("-unc",Arg.Set flags.unc,"Decide UNC property instead of confluence.");
  ("-uncsl",Arg.Set flags.uncsl,"Decide UNC property instead of confluence (shallow, left-linear).");
  ("-unr",Arg.Set flags.unr,"Decide UNR property rather than confluence.");
  ("-verbose",Arg.Set flags.vrb,"Be noisy.");
 ] in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,List.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let (>>=) = M.(>>=);;

(* Flatten a TRS R0.
 * Result is (eq, n, rs).
 * Subterms of the original TRS are mapped to integers 0 to n-1.
 * The flattened TRS has an equational part eq and a rewriting part rs.
 * eq is a list of triples (u, v, w) representing rules (u o v) -> w.
 * rs is a list of pairs (u, v), each representing a rewrite rule u -> v.
 *
 * eq describes a mapping of subterms of R0 to its subterms; there is
 * a unique constant from 0 to n-1 representing each subterm of R0.
 *)
let flatten : Trs.t -> Signature.t -> (int * eqn * (int * int) list) =
fun trs trs_sig ->
  let
    (* custom state monad *)
    (>>=) f g = fun s -> let (s', r) = f s in g r s'
  in let
    (* assign numbers to function symbols, reusing previous assignments *)
    mkcon f (cons, apps, i) =
      try ((cons, apps, i), H.find_ f cons)
      with Not_found -> ((H.add_ f i cons, apps, i+1), i)
  in let
    (* assign numbers to applications of the 'apply' operator *)
    mkapp a b (cons, apps, i) =
      try let apps_a = H.find_ a apps in
        try let n = H.find_ b apps_a in
          ((cons, apps, i), n)
        with Not_found ->
          ((cons, H.replace_ a (H.add_ b i apps_a) apps, i+1), i)
      with Not_found ->
        ((cons, H.add_ a (H.singleton b i) apps, i+1), i)
  in let rec
    term = function
    | Term.Fun (f, []) -> mkcon f
    | Term.Fun (_, [a; b]) ->
      term a >>= fun a' ->
      term b >>= fun b' ->
      mkapp a' b'
    | _ -> failwith "Impossible."
  in let
    rule rule (s, rs) =
      (term (Rule.lhs rule) >>= fun l ->
      term (Rule.rhs rule) >>= fun r ->
      fun s -> (s, (l, r) :: rs)) s
  in let
    ((cons, apps, n), rs) =
      Trs.fold rule ((H.create 0, H.create 0, 0), []) trs
  in let
    eq = H.fold (fun u -> H.fold (fun v w eq -> (u, v, w) :: eq)) apps []
  in
  (* F.printf "G: %d / %d / %d\n" (H.size apps) (List.length rs) n; *)
  if !(flags.vrb) then (
    F.printf "*** FLATTENING ***\n";
    H.iter (fun f u -> F.printf "  %s -> %d\n" (Signature.find_fun_name f trs_sig) u) cons;
    List.iter (fun (u, v, w) -> F.printf "  %d o %d -> %d\n" u v w) eq;
    List.iter (fun (u, v) -> F.printf "  %d -> %d\n" u v) rs;
  );
  (n, eq, rs)
;;

(* Rewrite closure of a flattened TRS, ala Plaisted:
 *
 * We have u --> v in the result whenever u -->^* v in the original TRS.
 *
 * The result is a relation between the constants created by the flattening
 * step, i.e., between the subterms of the original TRS.
 *)
let rewrite_closure : (int * eqn * (int * int) list) -> rel =
fun (n, eq, rs) ->
  let
    (* rc will be the result *) 
    rc = Array.make_matrix n n false and
    (* l stores eq rules u o v -> w, indexed by u; r stores them indexed by v *)
    l = Array.make n [] and r = Array.make n []
  in let
    eqn (u, v, w) =
      l.(u) <- (v, w) :: l.(u);
      r.(v) <- (u, w) :: r.(v);
  in let rec
    (* add a rule u -> v *)
    add (u, v) =
      if not rc.(u).(v) then (
        rc.(u).(v) <- true;
        (* transitivity *)
        for w = 0 to n-1 do
          if rc.(w).(u) then add (w, v); (* w -> u -> v *)
          if rc.(v).(w) then add (u, w); (* u -> v -> w *)
        done;
        (* left monotonicity *)
        List.iter (fun (u1, u2) ->
          List.iter (fun (v1, v2) ->
            if rc.(u1).(v1) then add (u2, v2); (* u2 = u o u1 -> v o v1 = v2 *)
          ) l.(v);
        ) l.(u);
        (* right monotonicity *)
        List.iter (fun (u1, u2) ->
          List.iter (fun (v1, v2) ->
            if rc.(u1).(v1) then add (u2, v2); (* u2 = u1 o u -> v1 o v = v2 *)
          ) r.(v);
        ) r.(u);
      );
  in
  (* fill l and r *)
  List.iter eqn eq;
  (* ensure reflexivity *)
  for i = 0 to n-1 do add (i, i); done;
  (* process rules one by one *)
  List.iter add rs;
  if !(flags.vrb) then (
    F.printf "*** REWRITE CLOSURE ***\n";
    for i = 0 to n-1 do for j = 0 to n-1 do
      if rc.(i).(j) then F.printf "  %d -> %d\n" i j; done; done;
  );
  rc
;;

(* Compute congruence closure efficiently in O(n log(n)) time.
 *
 * cf. Nieuwenhuis et al., Fast congruence closure and extensions, IC 2005
 *)
let congruence_closure0 : (int * eqn * (int * int) list) -> int array =
fun (n, eqs, rs) ->
  let
    repr = Array.init n (fun i -> i) and
    class_list = Array.init n (fun i -> [i]) and
    use_list = Array.make n [] and
    use_len = Array.make n 0 and
    lookup = H.create 0
  in
  List.iter (fun ((a1, a2, a) as r) ->
    use_list.(a1) <- r :: use_list.(a1);
    use_len.(a1) <- use_len.(a1) + 1;
    use_list.(a2) <- r :: use_list.(a2);
    use_len.(a2) <- use_len.(a2) + 1;
    H.replace lookup (a1, a2) a
  ) eqs;
  let rec
    merge (a, b) =
      let a = repr.(a) and b = repr.(b) in
      if a = b then () else
      let (a, b) = if use_len.(a) < use_len.(b) then (a, b) else (b, a) in
      List.iter (fun a ->
        class_list.(b) <- a :: class_list.(b);
        repr.(a) <- b
      ) class_list.(a);
      class_list.(a) <- [];
      use_len.(b) <- use_len.(a) + use_len.(b);
      let pending = ref [] in
      List.iter (fun ((c1, c2, c) as r) ->
        let c1 = repr.(c1) and c2 = repr.(c2) in
        try let c' = H.find lookup (c1, c2) in
          pending := (c, c') :: !pending;
        with Not_found ->
          H.replace lookup (c1, c2) c;
          use_list.(b) <- r :: use_list.(b);
      ) use_list.(a);
      use_list.(a) <- [];
      List.iter merge !pending
  in
  List.iter merge rs;
  if !(flags.vrb) then (
    F.printf "*** CONGRUENCE CLOSURE ***\n";
    for i = 0 to n-1 do if repr.(i) = i then (
     F.printf "  [";
     for j = 0 to n-1 do
      if repr.(i) = repr.(j) then
       F.printf "%d," j;
     done;
     F.printf "\008] = %d\n" i;
    ) done;
  );
  repr
;;

(* Congruence closure, the rewrite closure of R and its inverse.
 *
 * We have u --> v in the result whenever u <-->^* v in the original TRS.
 *)
let congruence_closure : (int * eqn * (int * int) list) -> rel =
fun (n, eq, rs) ->
  let repr = congruence_closure0 (n, eq, rs) in
  Array.init n (fun i -> Array.init n (fun j -> repr.(i) = repr.(j)))
;;

(* Calculate top stabilizable sides.
 *
 * A side u o v or constant u is top-stabilizable if it can be rewritten to
 * an (rc + eq)-normal form that is not a constant in a sequence of
 * (cc + eq^-1)-steps.
 *)
let topstabilizable :
  (int * eqn * (int * int) list) -> rel -> rel -> rel =
fun (n, eq, _) rc cc ->
  let
    (* nf stores (rc + eq)-normal forms of the form u o v *)
    nf = Array.make_matrix n n true and
    (* res.(u).(v) is [w] if u o v -> w in eq *)
    res = Array.make_matrix n n [] and
    (* cls.(u) stores the cc-equivalence-class belonging to u *)
    cls = Array.make n [] and
    (* sides collects top-stabilizable sides *)
    sides = Array.make_matrix n n false and
    (* consts collects top-stabilizable constants *)
    consts = Array.make n false
  in let
    (* mark terms reducible by eq/rc. assumes that rc is transitive. *)
    eqn (u, v, w) =
      for u' = 0 to n-1 do
        if rc.(u').(u) then
          for v' = 0 to n-1 do
            if rc.(v').(v) then
              (* u' o v' ->_rc u o v' ->_rc u o v ->_eq w *)
              nf.(u').(v') <- false;
          done;
      done;
      res.(u).(v) <- w :: res.(u).(v);
  in let rec
    (* add top-stabilizable u o v to 'sides' *)
    add_side u v =
      if not sides.(u).(v) then (
        let cs = ref [] in
        List.iter (fun u' ->
          List.iter (fun v' ->
            (* any cc-equivalent side is also top-stabilizable *)
            sides.(u').(v') <- true;
            (* any eq/cc-predecessor of u o v is a top-stabilizable constant *)
            cs := res.(u').(v') @ !cs;
          ) cls.(v)
        ) cls.(u);
        List.iter add_const !cs;
      );
  and
    (* add top-stabilizable u to 'consts' *)
    add_const u =
      if not consts.(u) then (
        consts.(u) <- true;
        for v = 0 to n-1 do
          (* u o v is top-stabilizable if either u or v is *)
          add_side u v;
          add_side v u;
        done;
      );
  in
  (* fill nf *)
  List.iter eqn eq;
  (* for u = 0 to n-1 do for v = 0 to n-1 do
    if nf.(u).(v) then F.printf "NF %d o %d\n" u v; done; done; *)
  (* fill cls *)
  for u = 0 to n-1 do
    for v = 0 to n-1 do
      if cc.(u).(v) then cls.(u) <- v :: cls.(u);
    done;
  done;
  (* fill sides and consts *)
  for u = 0 to n-1 do
    for v = 0 to n-1 do
      if nf.(u).(v) then add_side u v;
    done;
  done;
  if !(flags.vrb) then (
    F.printf "*** TOP-STABILIZABLE SIDES (old) ***\n";
    for i = 0 to n-1 do for j = 0 to n-1 do
      if sides.(i).(j) then F.printf "  %d o %d\n" i j; done; done;
  );
  sides
;;

(* Joinability for (eq^- + rc).
 *
 * Note that the code is very similar to the rewrite closure, with
 * a couple of significant differences:
 *
 * - rc is constant
 * - the direction of one transitivity check is different.
 *)
let joinable : int -> eqn -> rel -> rel =
fun n eq rc ->
  let
    (* join will be the result *) 
    join = Array.make_matrix n n false and
    (* l stores eq rules u o v -> w, indexed by u; r stores them indexed by v *)
    l = Array.make n [] and r = Array.make n []
  in let
    eqn (u, v, w) =
      l.(u) <- (v, w) :: l.(u);
      r.(v) <- (u, w) :: r.(v);
  in let rec
    (* add eq/rc-joinable pair (u, v) *)
    add u v =
      if not join.(u).(v) then (
        join.(u).(v) <- true;
        join.(v).(u) <- true;
        (* transitivity *)
        for w = 0 to n-1 do
          if rc.(w).(u) then add w v; (* w -> u -> ? <- v *)
          if rc.(w).(v) then add u w; (* u -> ? <- v <- w *)
        done;
        (* left monotonicity *)
        List.iter (fun (u1, u2) ->
          List.iter (fun (v1, v2) ->
            if join.(u1).(v1) then add u2 v2;
              (* u2 = u o u1 -> ? <- v o v1 = v2 *)
          ) l.(v);
        ) l.(u);
        (* right monotonicity *)
        List.iter (fun (u1, u2) ->
          List.iter (fun (v1, v2) ->
            if join.(u1).(v1) then add u2 v2;
              (* u2 = u1 o u -> ? <- v1 o v = v2 *)
          ) r.(v);
        ) r.(u);
      );
  in
  (* fill l and r *)
  List.iter eqn eq;
  (* ensure reflexivity *)
  for u = 0 to n-1 do add u u; done;
  join
;;

(* Finally, the confluence check.
 *)
let is_cr : (int * eqn * (int * int) list) -> bool = fun ((n, eq, rs) as fl) ->
  let rc = rewrite_closure fl in
  let cc = congruence_closure fl in
  let tops = topstabilizable fl rc cc in
  let
    (* representatives for the cc equivalence classes *)
    repr = Array.make n (-1) and
    (* top-stabilizable eq left-hand sides, by representative of rhs *)
    eq_repr = Array.make n [] and
    (* see 'joinable' *)
    join = joinable n eq rc
  in
  if !(flags.vrb) then (
    F.printf "*** JOINABLE CONSTANTS ***\n";
    for u = 0 to n-1 do for v = 0 to n-1 do
      if join.(u).(v) then F.printf "  %d \226\134\147 %d\n" u v; done; done;
  );
  (* fill in repr *)
  for u = 0 to n-1 do
    if repr.(u) < 0 then
      for v = u to n-1 do
        if cc.(u).(v) then
          repr.(v) <- u;
      done;
  done;
  (* fill in eq_repr *)
  List.iter (fun (u, v, w) ->
    if tops.(u).(v) then
      eq_repr.(repr.(w)) <- (u, v) :: eq_repr.(repr.(w));
  ) eq;
  try
    (* fun / fun condition:
     * All equivalent top-stabilizable sides have equivalent arguments.
     *)
    for w = 0 to n-1 do
      if not (List.is_empty eq_repr.(w)) then
        let (u, v) = List.hd eq_repr.(w) in
        List.iter (fun (u', v') ->
          if not cc.(u).(u') || not cc.(v).(v') then
            raise Exit;
        ) (List.tl eq_repr.(w));
    done;
    (* fun / con condition:
     * Every constant equivalent to a top-stabilizable side can be
     * rewritten to a side with equivalent arguments by eq^- + rc.
     *)
    for w0 = 0 to n-1 do
      List.iter (fun (u, v) ->
        for w = 0 to n-1 do
          if cc.(w0).(w) then
            if not (List.exists (fun (u', v', w') ->
              cc.(u).(u') && cc.(v).(v') && rc.(w).(w')) eq)
            then raise Exit;
        done;
      ) eq_repr.(w0);
    done;
    (* con / con condition:
     * all equivalent constants are joinable.
     *)
    for u = 0 to n-1 do
      for v = 0 to n-1 do
        if cc.(u).(v) && not join.(u).(v) then
          raise Exit;
      done;
    done;
    true
  with
    Exit -> false
;;

(* Compute normal form DFA
 *
 * The resulting automaton has n+2 stares (we need two catch-all states,
 * one for normal forms and one for reducible terms).
 * We also recover the set of constants of the original TRS and return
 * them as 'initial' states.
 *)
let mk_nf_aut : (int * eqn * (int * int) list) ->
  (bool array * bool array * (int -> int -> int)) =
fun (n, eq, rs) ->
  let finals = Array.make (n+2) true in
  (* we have two extra states: n is [*] and n+1 is a catch-all state for
   * reducible terms *)
  finals.(n+1) <- false;
  let inits = Array.make (n+2) true in
  inits.(n) <- false;
  inits.(n+1) <- false;
  let trans0 = H.create 0 in
  (* compute non-final states (reducible terms); the main observation is that
   * if a1 or a2 represents a reducible term and a1 o a2 -> a, then a
   * represents a reducible term as well *)
  let aux = Array.make n [] in
  List.iter (fun (a1,a2,a) ->
    aux.(a1) <- a :: aux.(a1);
    aux.(a2) <- a :: aux.(a2);
    H.replace trans0 (a1,a2) a;
    inits.(a) <- false;
  ) eq;
  let rec faux i = if not finals.(i) then () else (
    finals.(i) <- false;
    List.iter faux aux.(i)
  ) in
  List.iter (fun (l, r) -> faux l) rs;
  (* define transition function *)
  let trans a1 a2 =
    try let a = H.find trans0 (a1, a2) in a
    with Not_found -> if finals.(a1) && finals.(a2) then n else n+1 in
(*F.printf "NF finals:";
  for i = 0 to n+1 do if finals.(i) then F.printf " %d" i; done;
  F.printf "@."; *)
(*F.printf "NF inits:";
  for i = 0 to n+1 do if inits.(i) then F.printf " %d" i; done;
  F.printf "@."; *)
  (inits, finals, trans)
;;

(* Helper for checking UNC property, also used for NFP.
 *
 * This auxiliary function takes callbacks for each found normal form:
 * `add_con i` is called for constants, where `i` names the constant,
 * while `add_app (a1,a2,a)` is called for each a1 o a2 -> a in the
 * congruence closure such that a has a normal form, where a1, a2, and a
 * are representatives of a congruence class.
 *)
let is_unc_nfp_aux :
  (int -> unit) -> ((int * int * int) -> unit) ->
  (int * eqn * (int * int) list) -> bool =
fun add_con add_app ((n, eq, rs) as fl) ->
  let repr = congruence_closure0 fl in
  (* find (deterministic) quotient automaton C for convertible terms *)
  let eq' = List.unique_hash (List.map (
    fun (a1,a2,a) -> (repr.(a1), repr.(a2), repr.(a))) eq) in
  if !(flags.vrb) then List.iter (fun (u,v,w) -> F.printf "  %d o %d -> %d\n" u v w) eq';
  (* find deterministic automaton N for the normal forms *)
  let (nf_inits, nf_finals, nf_trans) = mk_nf_aut (n, eq, rs) in
  (* bookkeeping: for each state of C, record the transitions that contain
   * the state in the left-hand side *)
  let use_list = Array.make n [] in
  List.iter (fun ((a1,a2,a) as r) ->
    use_list.(a1) <- r :: use_list.(a1);
    if a1 != a2 then use_list.(a2) <- r :: use_list.(a2);
  ) eq';
  if !(flags.vrb) then F.printf "*** UNC CONDITION ***\n";
  let nf_state = Array.make n (-1) in
  (* now we enumerate all normal forms that are accepted by C in a bottom-up
   * fashion, stopping whenever a state of C accepts more than one normal form.
   * whenever such a normal form is found, the state in which N accepts it is
   * recorded in the nf_state array (which is indexed by states of C). *)
  let rec mark q qnf =
    if !(flags.vrb) then F.printf "  ? ->* (%d, %d)\n" q qnf;
    if nf_state.(q) > -1 then raise Exit (* second convertible normal form *)
    else (
      nf_state.(q) <- qnf;
      let us = List.filter (fun (a1,a2,a) ->
        nf_state.(a1) > -1 && nf_state.(a2) > -1) use_list.(q) in
      (* note: it is crucial that the 'us' list is computed before we do any
       * recursive calls, because they may change the nf_state array, and
       * then we risk enumerating the same normal form twice. *)
      List.iter (fun ((a1,a2,a) as e) ->
        let anf = nf_trans nf_state.(a1) nf_state.(a2) in
        if nf_finals.(anf) then (
          add_app e;
          mark a anf;
        )
      ) us
    ) in
  (* base case: loop over the normal form constants of the TRS *)
  for i = 0 to n-1 do
    if nf_inits.(i) && nf_finals.(i) then (
      add_con i;
      mark repr.(i) i;
    )
  done;
  true
;;

(* Check unique normal forms with respect to conversions (UNC) property.
 *)
let is_unc : (int * eqn * (int * int) list) -> bool =
fun fl ->
  try
    is_unc_nfp_aux (fun _ -> ()) (fun _ -> ()) fl
  with Exit -> false
;;

(* Check normal form property (NFP).
 *)
let is_nfp : (int * eqn * (int * int) list) -> bool =
fun ((n, eq, rs) as fl) ->
  let repr = congruence_closure0 fl in
  let rc = rewrite_closure (n, eq, rs) in
  let cc = congruence_closure fl in
  let tops = topstabilizable fl rc cc in
  let add_con a = (
    (* adding a constant normal form: all terms convertible to the constant
     * must reach the constant *)
    for i = 0 to n-1 do
      if repr.(i) = repr.(a) && not rc.(i).(a) then (
        if !(flags.vrb) then F.printf "  failed 1st NFP-condition for %d\n" a;
        raise Exit;
      );
    done;
    (* check that there are no corresponding top-stabilizable sides *)
    List.iter (fun (u1, u2, u) ->
      if repr.(u) = repr.(a) && tops.(u1).(u2) then (
        if !(flags.vrb) then F.printf "  failed 2nd NFP-condition for %d\n" a;
        raise Exit;
      );
    ) eq;
  )
  and add_app (a1, a2, a) =
    (* adding an application: must reach u1 o u2 congruent to a1 o a2
     * using forward steps and a (inverse) equation from eq *)
    (* u is "good" if u1 o u2 congruent to a1 o a2 and u1 o u2 -> u in E *)
    let good = (* List.map (fun (_, _, u) -> u) *)
      (List.filter (fun (u1, u2, _) -> repr.(u1) = a1 && repr.(u2) = a2) eq) in
    (* check that any i congruent to a reaches a "good" u *)
    for i = 0 to n-1 do
      if repr.(i) = a && List.for_all (fun (_, _, g) -> not rc.(i).(g)) good
      then (
        if !(flags.vrb) then F.printf "  failed 3rd NFP-condition for %d o %d -> %d\n" a1 a2 a;
        raise Exit;
      );
    done;
    (* check that for top-stabilizable sides, the arguments are in the
     * correct congruence class. *)
    List.iter (fun (u1, u2, u) ->
      if repr.(u) = repr.(a) && tops.(u1).(u2) &&
        (repr.(u1) != a1 || repr.(u2) != a2) then (
        if !(flags.vrb) then F.printf "  failed 4th NFP-condition for %d o %d -> %d\n" a1 a2 a;
        raise Exit;
      );
    ) eq;
    ()
  in
  try
    is_unc_nfp_aux add_con add_app fl
  with Exit -> false
;;

(* Check unique normal forms with respect to reductions (UNR) property.
 *)
let is_unr : (int * eqn * (int * int) list) -> bool =
fun (n, eq, rs) -> try
  (* compute /inverse/ rewrite closure *)
  let rc' = rewrite_closure (n, eq, List.map (fun (l,r) -> (r,l)) rs) in
  (* compute normal form automaton *)
  let (nf_inits, nf_finals, nf_trans) = mk_nf_aut (n, eq, rs) in
  (* keep track of terms with perfect sharing *)
  let idx = ref n and apps = H.create 0 in
  let mkidx a b = (let i = !idx in
    idx := i+1; if !(flags.vrb) then F.printf "  %d := %d o %d \n" i a b; i) in
  let mkapp a b =
    try let apps_a = H.find_ a apps in
      try let n = H.find_ b apps_a in n
      with Not_found -> (let i = mkidx a b in H.add apps_a b i; i)
    with Not_found -> (let i = mkidx a b in H.add apps a (H.singleton b i); i) in
  (* precompute lists of E-rules that mention a given state on their lhs *)
  let use_list = Array.make n [] in
  let use_list1 = Array.make n [] and use_list2 = Array.make n [] in
  List.iter (fun ((a1, a2, a) as r) ->
    use_list.(a1) <- r :: use_list.(a1);
    use_list.(a2) <- r :: use_list.(a2);
    use_list1.(a1) <- r :: use_list1.(a1);
    use_list2.(a2) <- r :: use_list2.(a2);
  ) eq;
  (* compute witnesses w(q) of normal forms that reach q by E \cup F^-.
   * Basically we enumerate runs in the product automaton of the normal form
   * automaton and the one defined by E \cup F^-. Unlike the is_unc function
   * above, that automaton is not deterministic, so we have to keep track of
   * the actual terms and compare them before declaring a clash.
   *)
  if !(flags.vrb) then F.printf "*** FIRST UNR-CONDITION ***\n";
  let nf_witness = Array.make n (-1) and nf_state = Array.make n (-1) in
  let rec mark q qnf w =
    if not nf_finals.(qnf) then () (* not a normal form *) else (
    if !(flags.vrb) then F.printf "  %d in W(%d, %d)\n" w q qnf;
    if nf_witness.(q) >= 0 then
      if nf_witness.(q) = w then () (* same NF *) else raise Exit (* not UNR *)
    else (
      nf_witness.(q) <- w;
      nf_state.(q) <- qnf;
      for i = 0 to n-1 do
        if rc'.(q).(i) then mark i qnf w; (* note: q -> i in F^- *)
      done;
      List.iter (fun (a1,a2,a) ->
        let w1 = nf_witness.(a1) and w2 = nf_witness.(a2) in
        if w1 >= 0 && w2 >= 0 then
          mark a (nf_trans nf_state.(a1) nf_state.(a2)) (mkapp w1 w2);
      ) use_list.(q);
    )) in
  for i = 0 to n-1 do
    (* base case: reflexivity *)
    if nf_inits.(i) then mark i i i;
  done;
  (* compute meetable constants as joinable constants w.r.t. E^- \cup F^-1 *)
  let meet = joinable n eq rc' in
  if !(flags.vrb) then (
    F.printf "*** MEETABLE CONSTANTS ***\n";
    for u = 0 to n-1 do for v = 0 to n-1 do
      if meet.(u).(v) then F.printf "  %d \226\134\145 %d\n" u v; done; done;
  );
  (* compute witnesses w'(q,qnf) that reach q via E \cup F^- steps, followed
   * by \uparrow, followed by E \cup F steps, and are accepted in qnf by the
   * normal form automaton.
   *)
  if !(flags.vrb) then F.printf "*** SECOND UNR-CONDITION ***\n";
  let nf_witness' = Array.make_matrix n (n+2) (-2) in
  let mkapp' a b = if a < 0 || b < 0 then -1 else mkapp a b in
  let rec mark' q qnf w =
    if not nf_finals.(qnf) then () (* not a normal form *) else (
    if !(flags.vrb) then F.printf "  %d in W'(%d, %d)\n" w q qnf;
    let wa = nf_witness'.(q).(qnf) in
    if wa = -1 || wa >= 0 && wa = w then () (* no change *)
    else (
      let wa' = if wa = -2 then w (* new *) else -1 (* clash *) in
      nf_witness'.(q).(qnf) <- wa';
      let wq = nf_witness.(q) in
      if wq >= 0 && wq != w then raise Exit; (* not UNR *)
      for i = 0 to n-1 do
        if rc'.(i).(q) then mark' i qnf w; (* note: q -> i in F *)
      done;
      List.iter (fun (a1,a2,a) ->
        for q2 = 0 to n+1 do
          let w2 = nf_witness'.(a2).(q2) in
          if w2 >= -1 then
            mark' a (nf_trans qnf q2) (mkapp' wa' w2);
         done;
      ) use_list1.(q);
      List.iter (fun (a1,a2,a) ->
        for q1 = 0 to n+1 do
          let w1 = nf_witness'.(a1).(q1) in
          if w1 >= -1 then
            mark' a (nf_trans q1 qnf) (mkapp' w1 wa');
         done;
      ) use_list2.(q);
    ))
  in
  for i = 0 to n-1 do
    if nf_witness.(i) >= 0 then
      for j = 0 to n-1 do
        (* base case: j \uparrow i with w(i) defined *)
        if meet.(i).(j) then mark' j nf_state.(i) nf_witness.(i);
      done;
  done;
  true;
with
  Exit -> false
;;

(* Unlike Trs.is_applicative, we also accept TRSs with constants only.
 *)
let is_applicative r =
  let rec check_sig b = function
   | [] -> M.return true
   | f::fs ->
    M.find_ari f >>= fun a ->
    if a = 0 then check_sig b fs
    else if a = 2 && not b then check_sig true fs else M.return false
  in
  check_sig false (Trs.funs r)
;;

let apply prop p =
  let trs = P.get_trs p in
  is_applicative trs >>= fun c ->
  M.get >>= fun trs_sig ->
  if not c || not (Trs.is_ground trs) then M.return None else
  let flat = flatten trs trs_sig in
  let r = prop flat in
  M.return (Some (if r then Result.make_yes [] else Result.make_no))
;;

let apply' prop p =
  let trs = P.get_trs p in
  prop trs >>= fun r ->
  M.return (Some (if r then Result.make_yes [] else Result.make_no))
;;

let solve fs p =
 let configure s = F.printf "%s@\n%!" s; flags.help := true in
 (try init (); Arg.parsex code spec fs with Arg.Bad s -> configure s);
 if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
 if P.is_crp p then (
  if !(flags.nfp) then apply is_nfp p else
  if !(flags.unc) then apply is_unc p else
  if !(flags.unr) then apply is_unr p else
  if !(flags.uncsl) then apply' ShallowLinUNC.is_unc p else
  apply is_cr p) >>= fun r ->
  M.return (Option.map (make p) r)
 else
  M.return None
;;

(* Compare Functions *)
let equal p q = P.equal (get_ip p) (get_ip q);;

(* Printers *)
let fprintf fs fmt p =
  if p.punr then
    if Result.is_yes (get_result p)
    then M.return (F.fprintf fmt "@[<1>%s:@\nUNR by decision procedure.@]" name)
    else M.return (F.fprintf fmt "@[<1>%s:@\nnot UNR by decision procedure.@]" name)
  else if p.punc || p.puncsl then
    if Result.is_yes (get_result p)
    then M.return (F.fprintf fmt "@[<1>%s:@\nUNC by decision procedure.%s@]" name (if p.puncsl then " (shallow, linear)" else ""))
    else M.return (F.fprintf fmt "@[<1>%s:@\nnot UNC by decision procedure.%s@]" name (if p.puncsl then " (shallow, linear)" else ""))
  else if p.pnfp then
    if Result.is_yes (get_result p)
    then M.return (F.fprintf fmt "@[<1>%s:@\nNFP by decision procedure.@]" name)
    else M.return (F.fprintf fmt "@[<1>%s:@\nnot NFP by decision procedure.@]" name)
  else  
    if Result.is_yes (get_result p)
    then M.return (F.fprintf fmt "@[<1>%s:@\nconfluent by decision procedure.@]" name)
    else M.return (F.fprintf fmt "@[<1>%s:@\nnon-confluent by decision procedure.@]" name)
;;

let fprintfx fs p =
  if p.punr then
    if Result.is_yes (get_result p)
    then XO.node "unrProof" [XO.node "groundUNR" []]
    else XO.node "unrDisproof" [XO.node "groundUNR" []]
  else if p.punc then
    if Result.is_yes (get_result p)
    then XO.node "uncProof" [XO.node "groundUNC" []]
    else XO.node "uncDisproof" [XO.node "groundUNC" []]
  else if p.puncsl then
    if Result.is_yes (get_result p)
    then XO.node "uncProof" [XO.node "shallowLinearUNC" []]
    else XO.node "uncDisproof" [XO.node "shallowLinearUNC" []]
  else if p.pnfp then
    if Result.is_yes (get_result p)
    then XO.node "nfpProof" [XO.node "groundNFP" []]
    else XO.node "nfpDisproof" [XO.node "groundNFP" []]
  else
    if Result.is_yes (get_result p)
    then XO.node "crProof" [XO.node "groundCR" []]
    else XO.node "crDisproof" [XO.node "groundCR" []]
;;
