(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Util;;
open Rewritingx;;
open Term;;

(*** MODULES ******************************************************************)
module C = Complexity;;
module F = Format;;
module M = Monad;;
module P = Problem;;
module XO = XmlOutput;;

(*** TYPES ********************************************************************)
type flags = {
  help  : bool ref;
  closed: bool ref;
  drop  : bool ref;
  m     : int ref;
  dup   : bool ref;
  redex : bool ref;
  star  : bool ref;
  force : bool ref;
  kb    : bool ref;
  okui  : bool ref;
  chew  : bool ref;
  p     : bool ref;
  rt    : bool ref;
  redundant : bool ref;
  cpcs  : bool ref;
  cpcs2 : bool ref;
  cpcscert : bool ref;
  ac    : bool ref;
 };;

type t = {
 ip : P.t;
 op : P.t;
 pm : int;
 pdup : bool;
 predex : bool;
 pstar : bool;
 pforce : bool;
 pkb : bool;
 pokui : bool;
 pchew : bool;
 pp : bool;
 prt : bool;
 predundant : bool;
 pcpcs : int;
 pac : bool;
};;

(*** GLOBALS ******************************************************************)
let code = "cr";;
let name = "Church Rosser Transformation Processor";;
let keywords = ["confluence transformation"; "church rosser transformation";"transformation"];;
let flags = {
 help  = ref false;
 drop  = ref false;
 closed= ref false;
 m     = ref 0;
 dup   = ref false;
 redex = ref false;
 star  = ref false;
 force  = ref false;
 kb    = ref false;
 okui  = ref false;
 chew  = ref false;
 p     = ref false;
 rt    = ref false;
 redundant = ref false;
 cpcs  = ref false;
 cpcs2 = ref false;
 cpcscert = ref false;
 ac    = ref false;
};;

let comment = "Transform Church Rosser problems in various ways.";;

let spec =
 let spec = [
  ("--help",Arg.Set flags.help,"Prints information about flags.");
  ("-help",Arg.Set flags.help,"Prints information about flags.");
  ("-h",Arg.Set flags.help,"Prints information about flags.");
  ("-closed",Arg.Set flags.closed,"prepares for closedness test of CPs.");
  ("-drop",Arg.Set flags.drop,"experimental: drop critical diagrams.");
  ("-m",Arg.Set_int flags.m,"search for (minimal+m)-length joins.");
  ("-dup",Arg.Set flags.dup,"prove relative termination of DUP.");
  ("-redex",Arg.Set flags.redex,"prove relative termination of REDEX.");
  ("-star",Arg.Set flags.star,"prove relative termination of STAR.");
  ("-force",Arg.Set flags.force,"force success of processor.");
  ("-kb",Arg.Set flags.kb,"return standard problem if CPs joinable.");
  ("-okui",Arg.Set flags.okui,"Okui's simultaneous critical pairs criterion.");
  ("-chew",Arg.Set flags.chew,"Chew's UNC criterion.");
  ("-p",Arg.Set flags.p,"print critical diagrams. (with -okui, " ^
   "prints simultameous critical peaks.)");
  ("-rt",Arg.Set flags.rt,"prepare relative termination phase.");
  ("-redundant",Arg.Set flags.redundant,"Filter redundant rules.");
  ("-cpcs",Arg.Set flags.cpcs,"Compute critical pair closing system (Thm 2.4).");
  ("-cpcs2",Arg.Set flags.cpcs2,"Compute critical pair closing system (Thm 2.11).");
  ("-cpcscert",Arg.Set flags.cpcscert,"Compute critical pair closing system using certifiable criteria.");
  ("-ac",Arg.Set flags.ac,"return AC termination problem if AC-CPs are joinable. " ^
   "(A standard problem with suitable symbols declared AC in the signature.)");
 ] in
 Arg.alignx 80 spec
;;

let help = (comment,keywords,List.map Triple.drop_snd spec);;

(*** FUNCTIONS ****************************************************************)
let make ?(n = ~-1) p q = Some {
  ip = p;
  pdup = !(flags.dup);
  predex = !(flags.redex);
  pstar = !(flags.star);
  pforce = !(flags.force);
  pkb = !(flags.kb);
  pokui = !(flags.okui);
  pchew = !(flags.chew);
  pm = !(flags.m);
  pp = !(flags.p);
  prt = !(flags.rt);
  predundant = !(flags.redundant);
  pcpcs =
    (if n < 0 then
       if !(flags.cpcs) then 1 else if !(flags.cpcs2) then 2 else n
     else n);
  pac = !(flags.ac);
  op = q;
};;
let failwithcode code s = failwith (Format.sprintf "%s: %s" code s);;
let (>>=) = M.(>>=);;
let init _ = 
 flags.help := false;
 flags.closed:= false;
 flags.drop := false;
 flags.m    := 0;
 flags.dup  := false;
 flags.redex := false;
 flags.star := false;
 flags.force := false;
 flags.kb   := false;
 flags.okui := false;
 flags.chew := false;
 flags.p    := false;
 flags.rt   := false;
 flags.redundant := false;
 flags.cpcs := false;
 flags.cpcs2 := false;
 flags.cpcscert := false;
 flags.ac := false;
;;

(* Destructors *)
let get_ip p = p.ip;;
let get_op p = p.op;;

let compute_dup trs = M.return (Trs.partition Rule.is_duplicating trs)

let of_pos rule pl pr = 
 M.map (Term.spine (Rule.lhs rule)) pl >>= fun cl ->
 M.map (Term.spine (Rule.rhs rule)) pr >>= fun cr ->
 M.return 
  (Trs.of_list (List.map (uncurry Rule.of_terms) (List.product cl cr)))
;;

let of_rule_x rule x = 
 let (pl,pr) = Rule.map (Term.subterm_pos x) rule in
 match pr with
  | [] -> 
   M.return (Trs.empty,Trs.empty)
  | [p] -> of_pos rule pl [p] >>= fun weak ->
   M.return (Trs.empty,weak)
  | pr -> of_pos rule pl pr >>= fun strict ->
   M.return (strict,Trs.empty)
;;

let merge (s0,w0) (s1,w1) = (Trs.union s0 s1,Trs.union w0 w1);;

let of_rule_star rule = 
 Rule.rename rule >>= fun rule ->
 let xl = List.map Term.make_var (Term.vars (Rule.lhs rule)) in
 let e = M.return (Trs.empty,Trs.empty) in
 List.foldr (fun x -> M.liftM2 merge (of_rule_x rule x)) e xl 
;;

let compute_star trs = 
 let e = M.return (Trs.empty,Trs.empty) in
 Trs.fold (fun rule acc -> M.liftM2 merge acc (of_rule_star rule)) e trs 
;;

let of_rule_redex rule = 
 let (l,r) = Rule.to_terms rule in
 let xs = Term.vars r in
 let dups = List.filter (fun x -> Term.count_var x l < Term.count_var x r) xs in
 (Trs.of_list (List.map (fun x -> Rule.of_terms l (Term.make_var x)) dups), Trs.singleton rule)
;;

let compute_redex trs = 
 let e = (Trs.empty,Trs.empty) in
 M.return (Trs.fold (fun rule acc -> merge acc (of_rule_redex rule)) e trs)
;;

let empty trs = M.return (Trs.empty,Trs.empty);;
let closed trs = M.return (trs, Trs.empty);;

let l cds trs = Trs.is_linear trs;;
let ll cds trs = Trs.is_left_linear trs;;
let ll_empty cds trs = Trs.is_left_linear trs && cds = [];;
let ll_trivial cds trs = 
 let cps = List.map Diagram.critical_pair cds in
 Trs.is_left_linear trs && List.for_all (uncurry Term.equal) cps
;;
let ttrue _ _ = true;;


(* filter redundant rules:
 * a rule l -> r is redundant if l -> r by some other rule in R *)
let filter_redundant trs =
  let rec check rs l r =
    Trs.rewrites l Position.root (Trs.of_list rs) r ||
    match (l,r) with
    | (Fun (f,xs),Fun (g,ys)) -> f = g && match_args rs xs ys
    | _ -> false
  and match_args rs xs ys = match (xs,ys) with
    | (x::xs,y::ys) ->
      if x = y then match_args rs xs ys else xs = ys && check rs x y
    | (xs   ,ys   ) -> xs = ys
  in let rec go rs rs' = match rs with
    | [] -> rs'
    | rl::rs -> let l = Rule.lhs rl and r = Rule.rhs rl in
      if check rs' l r || check rs l r then go rs rs' else go rs (rl::rs')
  in Trs.of_list (go (Trs.to_list trs) [])
;;

let apply g f p = 
 let s = P.get_strategy p and l = P.get_language p in
 let trs = P.get_trs p in
 let trs = if !(flags.redundant) then filter_redundant trs else trs in
 Diagram.critical_diagrams ~m:!(flags.m) trs >>= function
  | None -> M.return None (*not locally confluent*)
  | Some cds -> (
   if g cds trs then 
    f trs >>= fun (strict,weak) ->
    let cds = if !(flags.drop) (*|| !(flags.worth)*) then [] else cds in (*hack*)
    let q = P.make_ddp l s trs strict weak cds in
    let q = if !(flags.closed) then P.set_name "closed" q else q in
    M.return (make p q)
   else M.return None
   )
;;

let idem p = 
 let trs = P.get_trs p in
 Diagram.critical_diagrams ~m:!(flags.m) trs >>= function
  | None -> M.return None (*not locally confluent*)
  | Some _ -> M.return (make p (P.make_sp (P.get_language p) (P.get_strategy p) (P.get_trs p)))
;;

let is_f_c_rule f r = match Rule.to_terms r with
  | (Term.Fun (g, [Term.Var x1; Term.Var x2]), Term.Fun (h, [Term.Var y1; Term.Var y2])) ->
     Function.equal f g && Function.equal g h &&
     not (Variable.equal x1 x2) && Variable.equal x1 y2 && Variable.equal x2 y1
  | _ -> false
;;

let is_f_a_rule f r =
  match Rule.to_terms r with
  | (Term.Fun (f1, [Term.Fun (f2, [Term.Var x1; Term.Var x2]); Term.Var x3]),
     Term.Fun (g1, [Term.Var y1; Term.Fun (g2, [Term.Var y2; Term.Var y3])]))
  | (Term.Fun (g1, [Term.Var y1; Term.Fun (g2, [Term.Var y2; Term.Var y3])]),
     Term.Fun (f1, [Term.Fun (f2, [Term.Var x1; Term.Var x2]); Term.Var x3])) ->
     Function.equal f f1 &&
     Function.equal f1 f2 && Function.equal g1 g2 && Function.equal f1 g1 &&
     Variable.equal x1 y1 && Variable.equal x2 y2 && Variable.equal x3 y3 &&
     not (Variable.equal x1 x2) && not (Variable.equal x1 x3) && not (Variable.equal x2 x3)
 | _ -> false
;;

let fresh_ac_symbol f =
  M.find_ari f >>= fun a ->
  M.fresh_nary a >>= fun f' ->
  M.add_theory f' Theory.AC >>= fun _ ->
  M.return f'
;;

let make_ac trs =
  let fs = Trs.left_roots trs in
  let try_f_ac rs f =
    let (crs, rs') = List.partition (is_f_c_rule f) rs in
    let (ars, rs'') = List.partition (is_f_a_rule f) rs' in
    if List.is_empty crs || List.is_empty ars then
      M.return rs
    else
      if !(flags.p) then
        M.replace_theory f Theory.AC >>= fun _ ->
        M.return rs''
      else
        fresh_ac_symbol f >>= fun f' ->
        M.return (List.map (Rule.project (Term.map (fun g -> if g = f then f' else g))) rs'')
  in
  M.foldl try_f_ac (Trs.to_list trs) fs >>= fun rs ->
  M.return (Trs.of_list rs)
;;


let ac_wcr p =
  let trs = P.get_trs p in
  make_ac trs >>= fun trs ->
  Acrewrite.is_wcr trs >>= fun wcr ->
  if wcr || !(flags.p) then
    M.return (make p (P.make_sp (P.get_language p) (P.get_strategy p) trs))
  else
    M.return None
;;

let okui p =
 if !(flags.p) then M.return (make p p) else
 let trs = P.get_trs p in
 if not (Trs.is_left_linear trs) then M.return None else
 Trs.simultaneous_critical_pairs trs trs >>= fun scps ->
 if List.for_all (fun (l,r) ->
  not (List.is_empty (List.intersect (Rewrite.multistep trs l) (Rewrite.multistep trs r)))) scps
 then M.return (make p (P.set_sw Trs.empty Trs.empty p))
 else M.return None
;;

(* Chew's theorem *)
let chew_check_rhs m =
 let m' = function
  | Term.Var v -> m v
  | t -> [t] in
 let rec go t t' = Option.map_first (fun s ->
  Option.map_first (fun s' -> match (s, s') with
   | (Term.Var x, Term.Var y) -> if x = y then Some s else None
   | (Term.Fun (f, ts), Term.Fun (g, ts')) ->
    if f != g || List.length ts != List.length ts' then None else
    Option.map (fun ts'' -> Fun (f, ts''))
    (Option.map_all (fun (t,t') -> go t t') (List.zip ts ts'))
   | _ -> None
  ) (m' t')
 ) (m' t) in
 go
;;

let chew_check_overlay (rl1, l1', s1, r1) (rl2, l2', s2, r2) =
 try 
  let s = Elogic.unify l1' l2' in
  let m = List.map (fun (v, Term.Var v') ->
   (v', Substitution.apply_term s (Term.Var v)))
   (Substitution.to_list s1 @ Substitution.to_list s2) in
  let m' = fun v ->
   let r = List.filter (fun (v',t) -> v' = v) m in
   if r = [] then [Term.Var v] else List.map snd r in
  Option.map (fun x -> [(rl1, rl2, l1', l2', s, x)]) (chew_check_rhs m' r1 r2)
 with Elogic.Not_unifiable -> Some []
;;

let chew_check_overlays p =
 let rec tails' = function | [] -> [] | x::xs -> (x,xs) :: tails' xs in
 let trs = P.get_trs p in
 M.map (fun r ->
  Term.ren (Rule.lhs r) >>= fun l -> M.return (Rule.of_terms l (Rule.rhs r))
 ) (Trs.to_list trs) >>= fun trs' ->
 let xs = List.map (fun (r, r') ->
   let s = Elogic.match_term (Rule.lhs r) (Rule.lhs r') in
   (r, Rule.lhs r', s, Rule.rhs r')
  ) (List.zip (Trs.to_list trs) trs') in
 M.return (Option.map List.concat (Option.map_all (fun (a, xs') ->
  Option.map List.concat (Option.map_all (fun b ->
   chew_check_overlay a b) xs')) (tails' xs)))
;;

let chew p =
 let trs = P.get_trs p in
 M.map (fun r ->
  Term.ren (Rule.lhs r) >>= fun l -> M.return (Rule.of_terms l (Rule.rhs r))
 ) (Trs.to_list trs) >>= fun trs' ->
 Trs.overlaps ~outer:false (Trs.of_list trs') (Trs.of_list trs') >>= fun os ->
 if os != [] then M.return None else
 chew_check_overlays p >>= fun r ->
 M.return (if Option.is_some r then make p (P.set_sw Trs.empty Trs.empty p)
  else None)
;;

(* critical pair closing systems (Oyamaguch&Hirokawa IWC 2014) *)
(* Theorem 2.4 is implemented. TODO: try Theorem 2.11 as well *)
let cpcs p =
  let trs = P.get_trs p in
  let trs = if !(flags.redundant) then filter_redundant trs else trs in
  let rec select_step step =
    let rule = Rewrite.step_get_rule step in
    if !(flags.cpcscert) then M.return rule else
      let l = Rewrite.step_get_redex step in
      let s = Elogic.match_term l (Rule.lhs rule) in
      M.map (fun (v, t) -> Trs.tcap t trs >>= fun t -> M.return (v, t))
        (Substitution.to_list s) >>= fun s ->
      let s = Substitution.of_list s in
      M.return (Rule.of_terms (Substitution.apply_term s (Rule.lhs rule))
        (Substitution.apply_term s (Rule.rhs rule)))
  in let select_join js =
    (* simple heutristic: select join with smallest CPCs *)
    M.map (fun join -> M.map select_step (fst join @ snd join) >>= fun cpcs ->
      M.return (join, Trs.to_list (Trs.variant_free (Trs.of_list cpcs)))) js >>= fun js' ->
    M.return (List.foldl1 (fun t u -> if List.length (snd t) < List.length (snd u) then t else u) js')
  in let rec select_joins = function
  | [] -> M.return ([], [])
  | cd :: cds -> select_join (Diagram.joins cd) >>= function
    | (join, rs) -> select_joins cds >>= function
      | (cds, cpcs) ->
        M.return (Diagram.make (Diagram.peak cd) [join] :: cds, rs @ cpcs)
  in let rec is_multi_step ps = function
    | [] -> true
    | s :: ss ->
      let p = Rewrite.step_get_pos s and r = Rewrite.step_get_rule s in
      List.for_all (fun q ->
        Position.are_parallel p q ||
        Position.is_prefix p q &&
        not (Term.is_fun_pos (Position.difference q p) (Rule.lhs r))
      ) ps && is_multi_step (p :: ps) ss
  in let multistep_diagram cd =
    if Position.is_root (Rewrite.step_get_pos (fst (Diagram.peak cd))) then cd else
    Diagram.make (Diagram.peak cd) (List.filter (is_multi_step [] @.@ fst) (Diagram.joins cd))
  in let rec is_parallel_step ps = function
    | [] -> true
    | s :: ss ->
      let p = Rewrite.step_get_pos s and r = Rewrite.step_get_rule s in
      List.for_all (fun q -> Position.are_parallel p q) ps && is_parallel_step (p :: ps) ss
  in let parstep_diagram cd =
    if Position.is_root (Rewrite.step_get_pos (fst (Diagram.peak cd))) then cd else
    Diagram.make (Diagram.peak cd) (List.filter (is_parallel_step [] @.@ fst) (Diagram.joins cd))
  in if not (Trs.is_left_linear trs) then M.return None else
    Diagram.critical_diagrams ~m:!(flags.m) trs >>= function
    | None -> M.return None
    | Some cds ->
      match if !(flags.cpcs2) then
        let d = if !(flags.cpcscert) then parstep_diagram else multistep_diagram in
        let cds' = List.map d cds in
        if List.exists (List.is_empty @.@ Diagram.joins) cds' then None else Some cds'
      else Some cds with
    | None -> M.return None
    | Some cds ->
      select_joins cds >>= function
      | (cds, cpcs) ->
        let cpcs = Trs.variant_free (Trs.of_list cpcs) in
        if !(flags.cpcs) then
          let cpcs' = Trs.filter Rule.is_duplicating cpcs in
          let qcr = P.make_crp (P.get_language p) (P.get_strategy p) cpcs in
          let q = P.make_aux cpcs' trs qcr in
          if Trs.is_variant cpcs trs then M.return None else M.return (make p q)
        else
          let q = P.make_sp P.All P.Full cpcs in
          let n =
            if List.is_empty cds then 0 else
              List.flat_map Diagram.joins cds
              |> List.rev_unzip
              |> uncurry List.append
              |> List.map List.length
              |> List.foldl1 max
          in
          M.return (make ~n:n p q)
;;

(* Processor *)
let solve fs p =
 let configurate s = F.printf "%s@\n%!" s; flags.help := true in
 (try init (); Arg.parsex code spec fs with Arg.Bad s -> configurate s);
 if !(flags.help) then (Arg.usage spec ("Options for "^code^":"); exit 0);
 if P.is_crp p && Trs.is_trs (P.get_trs p) then 
  if !(flags.rt) then
   let q = P.make_aux (P.get_trs p) Trs.empty p in
   M.return (make p q)
  else if !(flags.closed) then apply ttrue closed p
  else if !(flags.force) then apply ttrue empty p
  else if !(flags.kb) then idem p
  else if !(flags.okui) then okui p
  else if !(flags.chew) then chew p
  else if !(flags.redex) then apply ll compute_redex p 
  else if !(flags.star) then apply ll compute_star p 
  else if !(flags.dup) then apply ll compute_dup p
  else if !(flags.cpcs) then cpcs p
  else if !(flags.cpcs2) then cpcs p
  else if !(flags.ac) then ac_wcr p
  else M.return None
 else if P.is_ddp p && !(flags.rt) then
  (* FIXME: is this useful? *)
  let l = P.get_language p and s = P.get_strategy p in
  let (strict,weak) = P.get_sw p in
  let q = P.make_rp l s strict weak in
  M.return (make p q)
 else 
  M.return None
;;

(* Complexity Bounds *)
(* let complexity c _ = C.mul c C.constant;; *)

(* Compare Functions *)
let equal p q = P.equal (get_ip p) (get_ip q);;

(* Printers *)
let fprintf_flags fmt p = 
 if p.pdup then Format.fprintf fmt " (dup)";
 if p.predex then Format.fprintf fmt " (redex)";
 if p.pstar then Format.fprintf fmt " (star)";
 if p.pkb then Format.fprintf fmt " (kb)";
 if p.pokui then Format.fprintf fmt " (okui)";
 if p.pchew then Format.fprintf fmt " (chew)";
 if p.prt then Format.fprintf fmt " (to relative problem)";
 if p.predundant then Format.fprintf fmt " (no redundant rules)";
 if p.pcpcs = 1 then Format.fprintf fmt " (critical pair closing system, Thm 2.4)";
 if p.pcpcs = 2 then Format.fprintf fmt " (critical pair closing system, Thm 2.11)";
 if p.pac then Format.fprintf fmt " (ac)";
;;

let fprintfm_diagrams fmt p = 
 F.fprintf fmt "@\n@[<1>%scritical peaks: " (if p.pokui then "simultaneous " else if p.pac then "AC " else "");
 if p.pokui then (
  let trs = P.get_trs (get_ip p) in
  Trs.simultaneous_overlaps trs trs >>= fun ols ->
  M.iter (fun ((ls,_,p) as o) ->
   Trs.simultaneous_critical_peak o >>= fun (u,v,w) ->
   F.fprintf fmt "@\n";
   Term.fprintfm fmt u >>= fun _ ->
   F.fprintf fmt "@\n <-[%a]- " (List.fprintf Position.fprintf "|")
    (List.map snd ls);
   Term.fprintfm fmt v >>= fun _ ->
   F.fprintf fmt " -[%a]->@\n" Position.fprintf p;
   Term.fprintfm fmt w >>= fun _ ->
   F.fprintf fmt "@\n"; M.return ()
  ) ols >>= fun _ ->
  M.return (if p.pp then F.fprintf fmt "@]" else F.fprintf fmt "joinable@]")
 ) else if p.pac && p.pp then (
   let trs = P.get_trs (get_op p) in
   Acrewrite.cpeaks trs >>= fun cps ->
   M.iter (fun  (u,v,w) ->
    Format.fprintf fmt "@\n@["; Term.fprintfm fmt u >>= fun _ ->
    Format.fprintf fmt "@ <-@ "; Term.fprintfm fmt v >>= fun _ ->
    Format.fprintf fmt "@ ->@ "; Term.fprintfm fmt w >>= fun _ ->
    Format.fprintf fmt "@]"; M.return ()
   ) cps >>= fun _ ->
   let peaks = List.length cps in
   M.map (fun (u,_,w) ->
     Acrewrite.ac_normalize u >>= fun u ->
     Acrewrite.ac_normalize w >>= fun w ->
     M.return (Rule.of_terms u w)) cps >>= fun cps ->
   let cps = Trs.variant_free (Trs.of_list cps) in
   F.fprintf fmt "@]@\n@[<1>AC critical pairs (%d/%d):@\n" (Trs.size cps) peaks;
   Trs.fprintfm fmt cps >>= fun _ ->
   M.return (F.fprintf fmt "@]")
 ) else if P.is_sp (get_op p) then M.return (F.fprintf fmt "joinable@]")
 else if P.is_ddp (get_op p) then
  let cds = Problem.get_cds (get_op p) in
  F.fprintf fmt "%d" (List.length cds);
  M.iter (fun d -> Format.fprintf fmt "@\n";
   if p.pp then Diagram.fprintfm fmt d
   else Diagram.fprintfm_peak fmt d) cds >>= fun _ ->
  F.fprintf fmt "@]";
  M.return ()
 else M.return ()
;;

let fprintfm_chew fmt p =
 chew_check_overlays (get_ip p) >>= fun (Some os) ->
 M.iter (fun (rl1, rl2, l1', l2', s, x) ->
  F.fprintf fmt "@[<1>The clusters for@\nr1 = ";
  Rule.fprintfm fmt rl1 >>= fun _ ->
  F.fprintf fmt "@\nr2 = ";
  Rule.fprintfm fmt rl2 >>= fun _ ->
  F.fprintf fmt "@]@\n@[<1>have a trivial peak:@\n";
  Term.fprintfm fmt x >>= fun _ ->
  F.fprintf fmt "@ <-@ ";
  Term.fprintfm fmt (Substitution.apply_term s l1') >>= fun _ ->
  F.fprintf fmt "@ ->@ ";
  Term.fprintfm fmt x >>= fun _ ->
  F.fprintf fmt "@]@\n"; M.return ()
 ) os >>= fun _ ->
 F.fprintf fmt "UNC by Chew's theorem."; M.return ()
;;

let fprintf fs fmt p =
 F.fprintf fmt "@[<1>%s%a:@\n" name fprintf_flags p; 
 P.fprintfm fmt (get_op p) >>= fun _ ->
 (if p.pchew then fprintfm_chew fmt p else fprintfm_diagrams fmt p) >>= fun _ ->
 F.fprintf fmt "@\n"; List.hd fs fmt >>= fun _ ->
 M.return (F.fprintf fmt "@]")
;;

let fprintfx fs p =
  if P.is_sp (get_op p) then
    if p.pcpcs >= 0 then
      XO.node "crProof" [
        XO.node "criticalPairClosingSystem" [
         P.fprintfx (get_op p); List.hd fs; XO.int "nrSteps" p.pcpcs
      ]]
    else
      XO.node "crProof" [
        XO.node "crKB" [P.fprintfx (get_op p); List.hd fs]
      ]
  else if P.is_ddp (get_op p) then
    if p.predundant && not (Trs.equivalent (P.get_trs (get_op p)) (P.get_trs (get_ip p))) then
      XO.node "crProof" [
        XO.node "redundantRules" [
          XO.node "trs" [Trs.fprintfx (P.get_trs (get_op p))];
          XO.int "nrSteps" 1;
          XO.node "crProof" [List.hd fs]]]
    else
      XO.node "crProof" [List.hd fs]
  else
    List.hd fs
;;
