(* Copyright 2008 Martin Korp, Christian Sternagel, Harald Zankl
 * GNU Lesser General Public License
 *
 * This file is part of TTT2.
 * 
 * TTT2 is free software: you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * TTT2 is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with TTT2. If not, see <http://www.gnu.org/licenses/>.
 *)

(*** OPENS ********************************************************************)
open Prelude;;

(*** INCLUDES *****************************************************************)

(*** MODULE TYPES *************************************************************)
module type ARG = sig
 type anon_fun = string -> unit
 type doc = string
 type key = string
 type spec =
  | Unit of (unit -> unit)
  | Bool of (bool -> unit)
  | Set of bool ref
  | Clear of bool ref
  | String of (string -> unit)
  | Set_string of string ref
  | Int of (int -> unit)
  | Set_int of int ref
  | Float of (float -> unit)
  | Set_float of float ref
  | Tuple of spec list
  | Symbol of string list * (string -> unit)
  | Rest of (string -> unit)
 type usage_msg = string
 
 exception Bad of string
 exception Help of string
 
 val current : int ref
 val parse : (key * spec * doc) list -> anon_fun -> usage_msg -> unit
 val usage : (key * spec * doc) list -> usage_msg -> unit
end

(*** TYPES ********************************************************************)
type error =
 | Unknown of string
 | Wrong of string * string * string
 | Missing of string
;;

type spec =
 | Unit of (unit -> unit)
 | Bool of (bool -> unit)
 | Set of bool ref
 | Clear of bool ref
 | String of (string -> unit)
 | Set_string of string ref
 | Int of (int -> unit)
 | Set_int of int ref
 | Float of (float -> unit)
 | Set_float of float ref
 | Tuple of spec list
 | Symbol of string list * (string -> unit)
 | Rest of (string -> unit)
;;

type key = Arg.key;;
type doc = Arg.doc;;
type anon_fun = Arg.anon_fun;;
type usage_msg = Arg.usage_msg;;

(*** EXCEPTIONS ***************************************************************)
exception Stop of error;;
exception Bad = Arg.Bad;;
exception Help = Arg.Help;;

(*** FUNCTIONS ****************************************************************)
let stop arg = function
 | Unknown s -> raise (Bad (Format.sprintf "%s: unknown option '%s'." arg s))
 | Missing s ->
  raise (Bad (Format.sprintf "%s: option '%s' needs an argument." arg s))
 | Wrong (o, a, e) ->
  raise (Bad (Format.sprintf
   "%s: wrong argument '%s'; option '%s' expects %s." arg a o e))
;;

let next a f failmsg = function
 | [] -> raise (Stop (Missing a))
 | arg :: args -> 
  try (f arg, args) 
  with
   | Invalid_argument _ -> raise (Stop (Wrong (a,arg,failmsg)))
   | Failure _ -> raise (Stop (Wrong (a,arg,failmsg)))
;;

let rec parsex arg spec = function
 | [] -> ()
 | a :: args ->
  let action =
   try Triple.snd (Listx.find ((=) a <.> Triple.fst) spec)
   with Not_found -> stop arg (Unknown a)
  in
  let process_action a args = function
   | Unit f -> f (); args
   | Bool f -> let (b,rest) = next a bool_of_string "a bool" args in f b; rest
   | Set b -> b := true; args
   | Clear b -> b := false; args
   | String f -> let (s,rest) = next a id "anything" args in f s; rest
   | Set_string s -> let (n,rest) = next a id "anything" args in s := n; rest
   | Int f -> let (n,rest) = next a int_of_string "an integer" args in f n; rest
   | Set_int i -> let (n,rest) = next a int_of_string "an integer" args in i := n; rest
   | Float f -> let (n,rest) = next a float_of_string "a float" args in f n; rest
   | Set_float f -> let (n,rest) = next a float_of_string "a float" args in f := n; rest
   | _ -> failwith "not supported"
  in
  let rest = try process_action a args action with Stop e -> stop arg e in
  parsex arg spec rest
;; 

let alignx w l =
 let max m (k,_,_) = let k = Stringx.length k in if k > m then k else m in
 let max_len = Listx.foldl max 0 l in
 Listx.map (fun (k,s,d) ->
  let msg = Stringx.align (w - (max_len + 3)) d in
  let lines = Stringx.split ~d:"[\n]" msg in
  let hl = Listx.hd lines in
  let hl = Stringx.shift_right (max_len - (Stringx.length k)) hl in
  let tl = Listx.tl lines in
  let tl = Listx.map (Stringx.shift_right (max_len + 3)) tl in
  let d = Listx.join id "\n" tl in
  (k, s, hl ^ if d = "" then "" else "\n" ^ d)) l
;;

let make_symlist prefix sep suffix l = match l with
 | [] -> "<none>"
 | h::t -> (Listx.foldl (fun x y -> x ^ sep ^ y) (prefix ^ h) t) ^ suffix
;;

let print_spec buf (key, spec, doc) = match spec with
 | Symbol (l,_) ->
  Printf.bprintf buf "  %s %s%s\n%!" key (make_symlist "{" "|" "}" l) doc
 | _ -> Printf.bprintf buf "  %s %s\n%!" key doc
;;

let usage_bx buf speclist errmsg =
 Printf.bprintf buf "%s\n%!" errmsg; Listx.iter (print_spec buf) speclist;
;;

let usagex fmt speclist errmsg =
 let b = Buffer.create 200 in
 usage_bx b speclist errmsg; Format.fprintf fmt "%s%!" (Buffer.contents b)
;;

(* This horrible conversion function exists so that the code compiles with
 * both ocaml 4.04 and 4.05, because the latter added a new constructor to
 * the spec type *)
let rec spec_to_spec = function
 | Unit s -> Arg.Unit s
 | Bool s -> Arg.Bool s
 | Set s -> Arg.Set s
 | Clear s -> Arg.Clear s
 | String s -> Arg.String s
 | Set_string s -> Arg.Set_string s
 | Int s -> Arg.Int s
 | Set_int s -> Arg.Set_int s
 | Float s -> Arg.Float s
 | Set_float s -> Arg.Set_float s
 | Tuple s -> Arg.Tuple (List.map spec_to_spec s)
 | Symbol (s,t) -> Arg.Symbol (s,t)
 | Rest s -> Arg.Rest s
;;

let spec_to_spec' = List.map (fun (k,s,d) -> (k,spec_to_spec s, d));;

let current = Arg.current;;
let usage s = Arg.usage (spec_to_spec' s);;
let parse s = Arg.parse (spec_to_spec' s);;
